/*
 * NAME:
 *	netio - interface stdio to read_net and write_net
 *
 * DESCRIPTION:
 *	These routines simply allow encryption hooks to be used with
 *	applications that use stdio. 
 *
 *	
 * AUTHOR:
 *	Simon J. Gerraty <sjg@quick.com.au>
 */
/*
 * RCSid:
 *	$Id: netio.h,v 1.6 1997/04/19 03:10:15 sjg Exp $
 *	
 *      @(#)Copyright (c) 1996 Simon J. Gerraty.
 *      
 *      This is free software.  It comes with NO WARRANTY.
 *      Permission to use, modify and distribute this source code 
 *      is granted subject to the following conditions.
 *      1/ that the above copyright notice and this notice 
 *      are preserved in all copies and that due credit be given 
 *      to the author.  
 *      2/ that any changes to this code are clearly commented 
 *      as such so that the author does not get blamed for bugs 
 *      other than his own.
 *      
 *      Please send copies of changes and bug-fixes to:
 *      sjg@quick.com.au
 */
#ifndef _NETIO_H_
#define _NETIO_H_

#include <sys/cdefs.h>
#ifndef _BSD_VA_LIST_ 
#if !defined(_STDARG_H) && !defined(_VARARGS_H)
#ifdef __STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif
#endif
#define _BSD_VA_LIST_ va_list
#endif

# define read		net_read
# define write		net_write
/*
 * If your system is pretty much like UNIX, then libsslfd
 * includes read,write,close calls which will ensure that your stdio
 * "just works", if things do not work, define NEED_NETIOB and
 * re-compile libsslfd
 */
#ifdef NEED_NETIOB
# define fgets		net_fgets
# define fgetln		net_fgetln
# define fprintf	net_fprintf
# define vfprintf	net_vfprintf
# define fflush		net_fflush
# define fclose		net_fclose
# undef putc
# undef getc
# undef ungetc
# define getc		net_fgetc
# define ungetc		net_ungetc
# define fgetc		net_fgetc
# define putc		net_fputc
# define fputc		net_fputc
#endif
#endif /* _NETIO_H_ */

__BEGIN_DECLS

int 	net_fflush	 __P((FILE * /* fp */));
int 	net_flush	 __P((int  /* fd */));
int 	net_fclose	 __P((FILE * /* fp */));
int 	net_vfprintf	 __P((FILE * /* fp */, const char * /* fmt */,  _BSD_VA_LIST_ /* ap */));
int 	net_fprintf	 __P((FILE * /* fp */, const char * /* fmt */, ...));
int 	net_fgetc	 __P((FILE * /* fp */));
int 	net_ungetc	 __P((int  /* c */, FILE * /* fp */));
int 	net_fputc	 __P((int  /* c */, FILE * /* fp */));
char *	net_fgets	 __P((char * /* buf */, int  /* nbytes */, FILE * /* fp */));
char *	net_fgetln	 __P((FILE * /* fp */, size_t * /* n */));
int 	net_read	 __P((int  /* fd */, unsigned char * /* buf */, int  /* nbytes */));
int 	net_write	 __P((int  /* fd */, unsigned char * /* buf */, int  /* nbytes */));

__END_DECLS
