/*
 *      @(#)Copyright (c) 1994, Simon J. Gerraty.
 *      
 *      This is free software.  It comes with NO WARRANTY.
 *      Permission to use, modify and distribute this source code 
 *      is granted subject to the following conditions.
 *      1/ that the above copyright notice and this notice 
 *      are preserved in all copies and that due credit be given 
 *      to the author.  
 *      2/ that any changes to this code are clearly commented 
 *      as such so that the author does not get blamed for bugs 
 *      other than his own.
 *      
 *      Please send copies of changes and bug-fixes to:
 *      sjg@quick.com.au
 */

#ifndef lint
static char RCSid[] = "$Id: sockdest.c,v 1.9 1998/03/22 21:49:10 sjg Exp $";
#endif
#include <sys/cdefs.h>
#include <stdio.h>
#ifdef __STDC__
#include <stdlib.h>
#endif
#include <sys/param.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <string.h>
#include <fcntl.h>
#include <ctype.h>
#include <errno.h>
#include <syslog.h>
#include <signal.h>
#ifdef SOCKS
# include "socks.h"
#endif

#if defined(_ANSI_SOURCE) || defined(_POSIX_SOURCE)
/* many systems omit these if the above defined */
char    *strdup __P((const char *));
#endif

/*
 * parse [{tcp,udp}!]dest[!port] or proto://host[:port]
 * into a sockaddr_in and a sock_type.
 * caller should set sock_type to a desired default if they don't like
 * SOCK_STREAM.
 */

char **
sock_dest(spec, sa, sock_type)
	const char *spec;
	struct sockaddr_in  *sa;
	int *sock_type;
{
	struct servent *sp = NULL;
	static char *a_list[2];
	extern char **get_addr_list();
	char *cp, *cp2, *dest = strdup(spec);
	char 	**addr_list = 0;
	int	i, stype, port;

	memset((char *) sa, 0, sizeof(struct sockaddr_in));

	spec = dest;
	stype = -1;
	port = 0;				/* any */

	if (cp = strstr(dest, "://")) {
		*cp = '\0';
		cp += 3;			/* now points to host */
		stype = SOCK_STREAM;
		if (cp2 = strchr(&cp[1], '/')) { 
			*cp2 = '\0';		/* lose path */
		}
		if (cp2 = strchr(&cp[1], ':')) {
			*cp2++ = '\0';
			port = atoi(cp2);
		} else {
			if ((sp = getservbyname(dest, "tcp")) != NULL ||
			    (sp = getservbyname(dest, "udp")) != NULL) {
				port = ntohs(sp->s_port);
				stype = (strcmp(sp->s_proto, "tcp") == 0) ?
					SOCK_STREAM : SOCK_DGRAM;
			} else if (strcmp(dest, "http") == 0) {
				port = 80;
			} else if (strcmp(dest, "https") == 0) {
				port = 443;
			} else if (strcmp(dest, "sshell") == 0) {
				port = 614;
			}
		}
		dest = cp;
	} else {
		if (dest[3] == ':' || dest[3] == '!') {
			if (strncmp(dest, "tcp", 3) == 0)
				stype = SOCK_STREAM;
			else if (strncmp(dest, "udp", 3) == 0)
				stype = SOCK_DGRAM;
		}
		if (stype > 0) {
			dest += 4;		/* past the : or ! */
		} else if (*sock_type <= 0)	/* no default set */
			stype = SOCK_STREAM;	/* default is tcp */
	}

	if (stype > 0)
		*sock_type = stype;
	
	if (port == 0 &&
	    (((i = strcspn(dest, ":!")) > 0) && dest[i] != '\0') ||
	    dest[0] == ':' ||
	    dest[0] == '!') {	/* port specified? */
		cp = &dest[i + 1];
		
		dest[i] = '\0';
		if (cp[0] == '*') {
			port = 0;
		} else {
			port = atoi(cp);
			if (port == 0 && *cp != '0') {	/* not a number? */
				char *proto = NULL;
				
				switch (*sock_type) {
				case SOCK_DGRAM:
					proto = "udp";
					break;
				case SOCK_STREAM:
					proto = "tcp";
					break;
				}
				if (proto != NULL &&
				    (sp = getservbyname(cp, proto)) != NULL) {
					port = ntohs(sp->s_port);
				} else {
					syslog(LOG_ERR, "sock_dest: unknown service '%s'", cp);
					goto out;
				}
			}
		}
	}
	
#ifdef BSD4_4
	sa->sin_len = sizeof(struct sockaddr_in);
#endif

	if (dest[0] == '\0' || dest[0] == '*') {
		sa->sin_family = AF_INET;
		sa->sin_addr.s_addr = INADDR_ANY;
		sa->sin_port = htons(port);
		a_list[0] = INADDR_ANY;
		a_list[1] = 0;
		addr_list = a_list;	/* just so we have non-null return */
	} else {
		addr_list = get_addr_list(dest);
		if (addr_list && *addr_list) {
			unsigned long ip_addr = *((unsigned long *) *addr_list);

			if (!(ip_addr == 0 || ip_addr == (unsigned long) -1L)) {
				sa->sin_family = AF_INET;
				sa->sin_addr.s_addr = ip_addr; /* in network order */
				sa->sin_port = htons(port);
			} else
				addr_list = 0;
		}
	}
 out:
	free((char *)spec);
	return (addr_list);		/* so caller can try other addresses */
}

#if defined(MAIN)
int
main(argc, argv)
	int argc;
	char **argv;
{
	char **addr_list;
	struct sockaddr_in  me;
	int errs = 0, i, sock_type;
	
	if (argc < 2)
		exit(1);

	while (*++argv) {
		sock_type = -1;
		if (!(addr_list = sock_dest(*argv, &me, &sock_type))) {
			errs++;
			fprintf(stderr, "invalid spec '%s'\n", *argv);
		} else {
			printf("dest: sock_type=%d, addr=%s, port=%d\n",
			       sock_type,
			       inet_ntoa(me.sin_addr),
			       ntohs(me.sin_port));
			for (i = 1; addr_list[i]; i++) {
				me.sin_addr.s_addr = *((unsigned long *) *addr_list);
				printf("\taddr=%s\n",
				       inet_ntoa(me.sin_addr));
			}
		}
	}
	exit(errs);
}
#endif
