/*
 * NAME:
 *	netiob - stdio like buffer management for netio
 *
 * DESCRIPTION:
 *	
 * AUTHOR:
 *	Simon J. Gerraty <sjg@quick.com.au>
 */
/*
 * RCSid:
 *	$Id: netiob.h,v 1.8 1998/11/05 07:37:25 sjg Exp $
 *	
 *      @(#)Copyright (c) 1996 Simon J. Gerraty.
 *      
 *      This is free software.  It comes with NO WARRANTY.
 *      Permission to use, modify and distribute this source code 
 *      is granted subject to the following conditions.
 *      1/ that the above copyright notice and this notice 
 *      are preserved in all copies and that due credit be given 
 *      to the author.  
 *      2/ that any changes to this code are clearly commented 
 *      as such so that the author does not get blamed for bugs 
 *      other than his own.
 *      
 *      Please send copies of changes and bug-fixes to:
 *      sjg@quick.com.au
 */
#ifndef _NETIOB_H_
#define _NETIOB_H_

#include <stdio.h>
#include <sys/cdefs.h>

#ifdef _STDIO_USES_IOSTREAM
/*
 * Red-Hat Linux - SOL
 */
# ifndef _IOB_FLAG
#   define NEED_NETIOB
#   ifdef _IO_NO_WRITES
#     define iswrite(fp) ((fp) && (((fp)->_flags & (_IO_NO_WRITES|_IO_CURRENTLY_PUTTING|_IO_IS_APPENDING)) == 0))
#   endif
# endif
#endif

#ifdef NEED_NETIOB
/*
 * Sadly, _iob internals are not as standard as they once were
 */
typedef struct net_iob {
	int   _cnt;				/* space left */
	unsigned char *_ptr;
	unsigned char *_base;
	int   _bufsiz;
	int   _flag;
	int   _file;
} NETIOB;

#define net_getc(p) (--(p)->_cnt < 0 ? net_filbuf(p) : (int)*(p)->_ptr++)
#define net_putc(x, p) (--(p)->_cnt < 0 ? \
			net_flsbuf((int)(x), (p)) \
                                : (int)(*(p)->_ptr++ = (x)))
#else
# define NETIOB FILE
# define net_getc(p) getc(p)
# define net_putc(x, p) putc((x), (p))
#endif

#ifndef _IOREAD
# define _IOFBF  0
# define _IOREAD 01
# define _IOWRT  02
# ifndef _IONBF
#   define _IONBF  04
# endif
# define _IOMYBUF        010
# define _IOEOF  020
# define _IOERR  040
# define _IOSTRG 0100
# ifndef _IOLBF
#   define _IOLBF  0200
# endif
# define _IORW   0400
#endif

/*
 * _flag is pretty consistent
 */
#ifndef _IOB_FLAG
# ifdef __SWR
#   define _IOB_FLAG _flags
#   define iswrite(fp) ((fp) && ((fp)->_IOB_FLAG & (__SWR)))
# else
#   define _IOB_FLAG _flag
# endif
#endif

#ifndef iswrite
# define iswrite(fp)	((fp) && ((fp)->_IOB_FLAG & (_IOWRT|_IORW)))
#endif
#ifndef NEW
# define NEW(x) (x *)malloc(sizeof(x))
#endif

#ifndef MIN
# define MIN(a, b) (((a) < (b)) ? (a) : (b))
#endif
#ifndef MAX
# define MAX(a, b) (((a) < (b)) ? (b) : (a))
#endif

#endif /* _NETIOB_H_ */

__BEGIN_DECLS

int     net_setvbuf      __P((NETIOB * /* p */, char * /* buf */, int  /* type */, size_t  /* size */));
NETIOB *        net_getiob       __P((int  /* fd */, int  /* force */, int /* w */));
NETIOB *        net_fgetiobt     __P((FILE * /* fp */, int (*/* test */)()));
NETIOB *        net_fgetiob      __P((FILE * /* fp */));
int 	_net_fflush	 __P((NETIOB * /* p */));
int     net_filbuf       __P((NETIOB * /* p */));
int     net_flsbuf       __P((int  /* c */, NETIOB * /* p */));
int 	net_fflush	 __P((FILE * /* fp */));
int 	net_flush	 __P((int  /* fd */));
int 	net_fclose	 __P((FILE * /* fp */));

__END_DECLS
