/*
 * NAME:
 *	rwhook - read/write hooks
 *
 * DESCRIPTION:
 *
 * AUTHOR:
 *	Simon J. Gerraty <sjg@quick.com.au>
 */
/*
 *      @(#)Copyright (c) 1996 Simon J. Gerraty.
 *      
 *      This is free software.  It comes with NO WARRANTY.
 *      Permission to use, modify and distribute this source code 
 *      is granted subject to the following conditions.
 *      1/ that the above copyright notice and this notice 
 *      are preserved in all copies and that due credit be given 
 *      to the author.  
 *      2/ that any changes to this code are clearly commented 
 *      as such so that the author does not get blamed for bugs 
 *      other than his own.
 *      
 *      Please send copies of changes and bug-fixes to:
 *      sjg@quick.com.au
 */
#ifndef lint
static char RCSid[] = "$Id: rwhook.c,v 1.3 1997/05/15 15:50:52 sjg Exp $";
#endif

#include <stdio.h>
#include "sslfd.h"


#ifndef NUMFDS
# define NUMFDS 64
#endif

#define FD_OK(fd) (fd >= 0 && fd < NUMFDS)

static RWHOOK hook[NUMFDS];

void *
get_hook(fd, flag)
	int fd;
	int flag;
{
	RWHOOK *hp;
	static int once = 1;

	if (once) {
		memset((char *)hook, 0, sizeof(hook));
		once = 0;
	}
	if (FD_OK(fd)) {
		hp = &hook[fd];
		if (hp->flag) {
			switch (flag) {
			case 0:	return hp;
			case HK_READ:	return hp->read;
			case HK_WRITE:	return hp->write;
			case HK_CLOSE:	return hp->close;
			}
		}
	}
	return 0;
}

int
set_hook(fd, flag, fp)
	int fd;
	int flag;
	int (*fp) ();
{
	RWHOOK *hp;
	static int once = 1;

	if (once) {
		(void) get_hook(-1, 0);
		once = 0;
	}
	if (FD_OK(fd)) {
		hp = &hook[fd];
		if (!flag && !fp) {
			memset((char *)hp, 0, sizeof(RWHOOK));
		} else {
			if (flag & HK_WRITE)
				hp->write = fp;
			if (flag & HK_CLOSE)
				hp->close = fp;
			if (flag & HK_READ)
				hp->read = fp;
			if (fp)
				hp->flag |= flag;
			else
				hp->flag &= ~flag;
		}
		return 1;
	}
	return 0;
}

